import sys
import warnings
from math import log
from numbers import Number

import numpy as np

from joblib import Parallel, delayed

from scipy.optimize import fmin_l_bfgs_b

from sklearn.base import clone
from sklearn.base import is_regressor
from sklearn.multioutput import MultiOutputRegressor
from sklearn.utils import check_random_state

from ..acquisition import _gaussian_acquisition
from ..acquisition import gaussian_acquisition_1D
from ..learning import GaussianProcessRegressor
from ..space import Categorical
from ..space import Space
from ..space.constraints import Constraints
from ..utils import check_x_in_space
from ..utils import cook_estimator
from ..utils import create_result
from ..utils import has_gradients
from ..utils import is_listlike
from ..utils import is_2Dlistlike
from ..utils import normalize_dimensions


class Optimizer(object):
    """Run bayesian optimisation loop.

    An `Optimizer` represents the steps of a bayesian optimisation loop. To
    use it you need to provide your own loop mechanism. The various
    optimisers provided by `ProcessOptimizer` use this class under the hood.

    Use this class directly if you want to control the iterations of your
    bayesian optimisation loop.

    Parameters
    ----------
    * `dimensions` [list, shape=(n_dims,)]:
        List of search space dimensions.
        Each search dimension can be defined either as

        - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
          dimensions),
        - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
          dimensions),
        - as a list of categories (for `Categorical` dimensions), or
        - an instance of a `Dimension` object (`Real`, `Integer` or
          `Categorical`).

    * `base_estimator` ["GP", "RF", "ET", "GBRT" or sklearn regressor, default="GP"]:
        Should inherit from `sklearn.base.RegressorMixin`.
        In addition the `predict` method, should have an optional `return_std`
        argument, which returns `std(Y | x)`` along with `E[Y | x]`.
        If base_estimator is one of ["GP", "RF", "ET", "GBRT"], a default
        surrogate model of the corresponding type is used corresponding to what
        is used in the minimize functions.

    * `n_random_starts` [int, default=10]:
        DEPRECATED, use `n_initial_points` instead.

    * `n_initial_points` [int, default=10]:
        Number of evaluations of `func` with initialization points
        before approximating it with `base_estimator`. Points provided as
        `x0` count as initialization points. If len(x0) < n_initial_points
        additional points are sampled at random.

    * `lhs` [bool, default = True]:
        If set to True, the optimizer will use latin hypercube sampling for the 
        first n_initial_points. If set to False, the optimizer will return 
        random points

    * `acq_func` [string, default=`"gp_hedge"`]:
        Function to minimize over the posterior distribution. Can be either

        - `"LCB"` for lower confidence bound.
        - `"EI"` for negative expected improvement.
        - `"PI"` for negative probability of improvement.
        - `"gp_hedge"` Probabilistically choose one of the above three
          acquisition functions at every iteration.
            - The gains `g_i` are initialized to zero.
            - At every iteration,
                - Each acquisition function is optimised independently to
                  propose an candidate point `X_i`.
                - Out of all these candidate points, the next point `X_best` is
                  chosen by $softmax(\eta g_i)$
                - After fitting the surrogate model with `(X_best, y_best)`,
                  the gains are updated such that $g_i -= \mu(X_i)$
        - `"EIps" for negated expected improvement per second to take into
          account the function compute time. Then, the objective function is
          assumed to return two values, the first being the objective value and
          the second being the time taken in seconds.
        - `"PIps"` for negated probability of improvement per second. The
          return type of the objective function is assumed to be similar to
          that of `"EIps

    * `acq_optimizer` [string, `"sampling"` or `"lbfgs"`, default=`"auto"`]:
        Method to minimize the acquistion function. The fit model
        is updated with the optimal value obtained by optimizing `acq_func`
        with `acq_optimizer`.

        - If set to `"auto"`, then `acq_optimizer` is configured on the
          basis of the base_estimator and the space searched over.
          If the space is Categorical or if the estimator provided based on
          tree-models then this is set to be "sampling"`.
        - If set to `"sampling"`, then `acq_func` is optimized by computing
          `acq_func` at `n_points` randomly sampled points.
        - If set to `"lbfgs"`, then `acq_func` is optimized by
              - Sampling `n_restarts_optimizer` points randomly.
              - `"lbfgs"` is run for 20 iterations with these points as initial
                points to find local minima.
              - The optimal of these local minima is used to update the prior.

    * `random_state` [int, RandomState instance, or None (default)]:
        Set random state to something other than None for reproducible
        results.

    * `acq_func_kwargs` [dict]:
        Additional arguments to be passed to the acquistion function.

    * `acq_optimizer_kwargs` [dict]:
        Additional arguments to be passed to the acquistion optimizer.
        options are:
        - "length_scale_bounds" [list] a list of tuples with lower and upper bound
        -  "length_scale" [list] a list of floats
        - "n_restarts_optimizer" [int]
        - "n_jobs" [int]
        
    * `n_objectives` [int, default=1]:
        Number of objectives to be optimized. 
        When n_objectives>1 the optimizer will fit models for each objective and the Pareto front can be approximated using NSGA2



    Attributes
    ----------
    * `Xi` [list]:
        Points at which objective has been evaluated.
    * `yi` [scalar]:
        Values of objective at corresponding points in `Xi`.
    * `models` [list]:
        Regression models used to fit observations and compute acquisition
        function.
    * `space`
        An instance of `ProcessOptimizer.space.Space`. Stores parameter search space used
        to sample points, bounds, and type of parameters.

    """

    def __init__(
        self,
        dimensions,
        base_estimator="gp",
        n_random_starts=None,
        n_initial_points=10,
        lhs=True,
        acq_func="gp_hedge",
        acq_optimizer="auto",
        random_state=None,
        acq_func_kwargs=None,
        acq_optimizer_kwargs=None,
        n_objectives=1,
    ):
        self.rng = check_random_state(random_state)

        # Set the number of objectives
        self.n_objectives = n_objectives

        # Configure acquisition function

        # Store and creat acquisition function set
        self.acq_func = acq_func
        self.acq_func_kwargs = acq_func_kwargs

        allowed_acq_funcs = ["gp_hedge", "EI", "LCB", "PI", "EIps", "PIps", ""]
        if self.acq_func not in allowed_acq_funcs:
            raise ValueError(
                "expected acq_func to be in %s, got %s"
                % (",".join(allowed_acq_funcs), self.acq_func)
            )

        # treat hedging method separately
        if self.acq_func == "gp_hedge":
            self.cand_acq_funcs_ = ["EI", "LCB", "PI"]
            self.gains_ = np.zeros(3)
        else:
            self.cand_acq_funcs_ = [self.acq_func]

        if acq_func_kwargs is None:
            acq_func_kwargs = dict()
        self.eta = acq_func_kwargs.get("eta", 1.0)

        # Configure counters of points

        # Check `n_random_starts` deprecation first
        if n_random_starts is not None:
            warnings.warn(
                (
                    "n_random_starts will be removed in favour of "
                    "n_initial_points."
                ),
                DeprecationWarning,
            )
            n_initial_points = n_random_starts

        if n_initial_points < 0:
            raise ValueError(
                "Expected `n_initial_points` >= 0, got %d" % n_initial_points
            )
        self._n_initial_points = n_initial_points
        self.n_initial_points_ = n_initial_points

        # record other arguments
        if acq_optimizer_kwargs is None:
            acq_optimizer_kwargs = dict()

        self._length_scale_bounds = acq_optimizer_kwargs.get(
            "length_scale_bounds", None
        )
        self._length_scale = acq_optimizer_kwargs.get("length_scale", None)
        self.n_points = acq_optimizer_kwargs.get("n_points", 10000)
        self.n_restarts_optimizer = acq_optimizer_kwargs.get(
            "n_restarts_optimizer", 5
        )
        n_jobs = acq_optimizer_kwargs.get("n_jobs", 1)
        self.n_jobs = n_jobs
        self.acq_optimizer_kwargs = acq_optimizer_kwargs

        # Configure estimator
        self._check_length_scale_bounds(dimensions, self._length_scale_bounds)
        # build base_estimator if doesn't exist
        if isinstance(base_estimator, str):
            base_estimator = cook_estimator(
                base_estimator,
                space=dimensions,
                random_state=self.rng.randint(0, np.iinfo(np.int32).max),
                length_scale_bounds=self._length_scale_bounds,
                length_scale=self._length_scale,
            )

        # check if regressor
        if not is_regressor(base_estimator) and base_estimator is not None:
            raise ValueError("%s has to be a regressor." % base_estimator)

        # treat per second acqusition function specially
        is_multi_regressor = isinstance(base_estimator, MultiOutputRegressor)
        if "ps" in self.acq_func and not is_multi_regressor:
            self.base_estimator_ = MultiOutputRegressor(base_estimator)
        else:
            self.base_estimator_ = base_estimator

        # Configure optimizer

        # decide optimizer based on gradient information
        if acq_optimizer == "auto":
            if has_gradients(self.base_estimator_):
                acq_optimizer = "lbfgs"
            else:
                acq_optimizer = "sampling"

        if acq_optimizer not in ["lbfgs", "sampling"]:
            raise ValueError(
                "Expected acq_optimizer to be 'lbfgs' or "
                "'sampling', got {0}".format(acq_optimizer)
            )

        if (
            not has_gradients(self.base_estimator_)
            and acq_optimizer != "sampling"
        ):
            raise ValueError(
                "The regressor {0} should run with "
                "acq_optimizer"
                "='sampling'.".format(type(base_estimator))
            )
        self.acq_optimizer = acq_optimizer

        # Configure search space

        # normalize space if GP regressor
        if isinstance(self.base_estimator_, GaussianProcessRegressor):
            dimensions = normalize_dimensions(dimensions)
        self.space = Space(dimensions)

        if (
            isinstance(self.base_estimator_, GaussianProcessRegressor)
            and self.space.is_categorical
        ):
            raise ValueError(
                "GaussianProcessRegressor on a purely categorical space"
                " is not supported. Please use another base estimator"
            )
        # Latin hypercube sampling

        self._lhs = lhs
        if lhs:
            self._lhs_samples = self.space.lhs(n_initial_points)

        # Default is no constraints
        self._constraints = None
        # record categorical and non-categorical indices
        self._cat_inds = []
        self._non_cat_inds = []
        for ind, dim in enumerate(self.space.dimensions):
            if isinstance(dim, Categorical):
                self._cat_inds.append(ind)
            else:
                self._non_cat_inds.append(ind)

        # Initialize storage for optimization

        self.models = []
        self.Xi = []
        self.yi = []

        # Initialize cache for `ask` method responses

        # This ensures that multiple calls to `ask` with n_points set
        # return same sets of points. Reset to {} at every call to `tell` and `set_constraints`.
        self.cache_ = {}

    def copy(self, random_state=None):
        """Create a shallow copy of an instance of the optimizer.

        Parameters
        ----------
        * `random_state` [int, RandomState instance, or None (default)]:
            Set the random state of the copy.
        """

        optimizer = Optimizer(
            dimensions=self.space.dimensions,
            base_estimator=self.base_estimator_,
            n_initial_points=self.n_initial_points_,
            lhs=self._lhs,
            acq_func=self.acq_func,
            acq_optimizer=self.acq_optimizer,
            acq_func_kwargs=self.acq_func_kwargs,
            acq_optimizer_kwargs=self.acq_optimizer_kwargs,
            random_state=random_state,
            n_objectives=self.n_objectives,
        )

        # It is important to copy the constraints so that a call to '_tell()' will create a valid _next_x
        optimizer._constraints = self._constraints
        if self._lhs:
            optimizer._lhs_samples = self._lhs_samples

        if hasattr(self, "gains_"):
            optimizer.gains_ = np.copy(self.gains_)

        if self.Xi:
            optimizer._tell(self.Xi, self.yi)

        return optimizer

    def ask(self, n_points=None, strategy="stbr_fill"):
        """Query point or multiple points at which objective should be evaluated.

        * `n_points` [int or None, default=None]:
            Number of points returned by the ask method.
            If the value is None, a single point to evaluate is returned.
            Otherwise a list of points to evaluate is returned of size
            n_points. This is useful if you can evaluate your objective in
            parallel, and thus obtain more objective function evaluations per
            unit of time.

        * `strategy` [string, default=`stbr_fill`]:
            Method to use to sample multiple points (see also `n_points`
            description). This parameter is ignored if n_points = None.
            Supported options are `"cl_min"`, `"cl_mean"` or `"cl_max"`.
            
            -if set to `"stbr_fill"` then Steinerberger sampling is used
              after first point.
            -if set to `"stbr_full"` then Steinerberger sampling is used
              from first point.   
              
             For details on the Steinerberger method see:
             https://arxiv.org/abs/1902.03269
             
            - If set to `"cl_min"`, then constant liar strategy is used
               with lie objective value being minimum of observed objective
               values. `"cl_mean"` and `"cl_max"` means mean and max of values
               respectively. For details on this strategy see:

               https://hal.archives-ouvertes.fr/hal-00732512/document

               With this strategy a copy of optimizer is created, which is
               then asked for a point, and the point is told to the copy of
               optimizer with some fake objective (lie), the next point is
               asked from copy, it is also told to the copy with fake
               objective and so on. The type of lie defines different
               flavours of `cl_x` strategies.

            - If set to `"KB"`, then Kriging believer is used. Kriging believer
                is returning the expected value of the next_x before asking for
                the next_next_x. Implementation is still experimental.
                For use of KB, see:
                https://www.nature.com/articles/s41586-021-03213-y

        """

        if not (
            (isinstance(n_points, int) and n_points > 0) or n_points is None
        ):
            raise ValueError(
                "n_points should be int > 0, got " + str(n_points)
            )
        # These are the only filling strategies which are supported
        supported_strategies = [
            "cl_min",
            "cl_mean",
            "cl_max",
            "stbr_fill",
            "stbr_full",
            "KB"
        ]

        if strategy not in supported_strategies:
            raise ValueError(
                "Expected parallel_strategy to be one of "
                + str(supported_strategies)
                + ", "
                + "got %s" % strategy
            )

        if strategy == "stbr_full" and self._n_initial_points < 1:

            # Steienerberger sampling can not be used from an empty Xi set
            if self.Xi == []:
                raise ValueError(
                    "Steinerberger sampling requires initial points but got [] "
                )

            if n_points is None:
                # Returns a single Steinerberger point
                X = self.stbr_scipy()
            else:
                # Returns 'n_points' Steinerberger points
                X = self.stbr_scipy(n_points=n_points)
            return X

        if n_points is None or n_points == 1:
            return self._ask()

        # Caching the result with n_points not None. If some new parameters
        # are provided to the ask, the cache_ is not used.
        if (n_points, strategy) in self.cache_:
            return self.cache_[(n_points, strategy)]

        # Copy of the optimizer is made in order to manage the
        # deletion of points with "lie" objective (the copy of
        # optimizer is simply discarded)
        opt = self.copy(
            random_state=self.rng.randint(0, np.iinfo(np.int32).max)
        )

        X = []
        for i in range(n_points):
            if (
                i > 0
                and strategy == "stbr_fill"
                and self._n_initial_points < 1
            ):
                x = opt.stbr_scipy()[0]
            else:
                x = opt._ask()
            X.append(x)

            ti_available = "ps" in self.acq_func and len(opt.yi) > 0
            ti = [t for (_, t) in opt.yi] if ti_available else None

            if strategy == "KB":
                reshaped_x = np.asarray(x).reshape(1, -1)
                transformed_x = opt.space.transform(reshaped_x.tolist())
                if opt.n_objectives == 1:
                    y_lie = opt.models[-1].predict(transformed_x)[0]
                else:
                    y_lie = []
                    for model in opt.models[-1]:
                        y_lie.append(model.predict(transformed_x)[0])

            elif strategy == "cl_min":
                y_lie = (
                    np.min(opt.yi, axis=0).tolist()
                    if opt.yi
                    else np.zeros(opt.n_objectives).tolist()
                )  # CL-min lie
                if opt.n_objectives == 1 and not opt.yi:
                    y_lie = y_lie[0]
                t_lie = (
                    np.min(ti) if ti is not None else log(sys.float_info.max)
                )
            elif strategy == "cl_mean":
                y_lie = (
                    np.mean(opt.yi, axis=0).tolist()
                    if opt.yi
                    else np.zeros(opt.n_objectives).tolist()
                )  # CL-mean lie
                if opt.n_objectives == 1 and not opt.yi:
                    y_lie = y_lie[0]
                t_lie = (
                    np.mean(ti) if ti is not None else log(sys.float_info.max)
                )
            else:
                y_lie = (
                    np.max(opt.yi, axis=0).tolist()
                    if opt.yi
                    else np.zeros(opt.n_objectives).tolist()
                )  # CL-max lie
                if opt.n_objectives == 1 and not opt.yi:
                    y_lie = y_lie[0]
                t_lie = (
                    np.max(ti) if ti is not None else log(sys.float_info.max)
                )

            # Lie to the optimizer.
            if "ps" in self.acq_func:
                # Use `_tell()` instead of `tell()` to prevent repeated
                # log transformations of the computation times.
                opt._tell(x, (y_lie, t_lie))
            else:
                opt._tell(x, y_lie)

        self.cache_ = {(n_points, strategy): X}  # cache_ the result

        return X

    def _ask(self):
        """Suggest next point at which to evaluate the objective.

        Return a random point while not at least `n_initial_points`
        observations have been `tell`ed, after that `base_estimator` is used
        to determine the next point.
        """

        if self._n_initial_points > 0 or self.base_estimator_ is None:
            # this will not make a copy of `self.rng` and hence keep advancing
            # our random state.

            assert not (
                self._constraints and self._lhs
            ), "Constraints can't be used while latin hypercube sampling is not exhausted"

            if self._n_initial_points == 0 and self.base_estimator_ is None:
                # This occurs during runs with dummy minimizer in which base_estimator is None by design
                return self.space.rvs(random_state=self.rng)[0]

            if self._constraints:
                # We use another sampling method when constraints are added
                return self._constraints.rvs(random_state=self.rng)[0]
            elif self._lhs:
                # The samples are evaluated starting form lhs_samples[0]
                return self._lhs_samples[
                    len(self._lhs_samples) - self._n_initial_points
                ]
            else:
                return self.space.rvs(random_state=self.rng)[0]
        else:
            if not self.models:
                raise RuntimeError(
                    "Random evaluations exhausted and no "
                    "model has been fit."
                )

            next_x = self._next_x

            min_delta_x = min(
                [self.space.distance(next_x, xi) for xi in self.Xi]
            )

            if abs(min_delta_x) <= 1e-8:
                warnings.warn(
                    "The objective has been evaluated " "at this point before."
                )

            # return point computed from last call to tell()
            return next_x

    def tell(self, x, y, fit=True):
        """Record an observation (or several) of the objective function.

        Provide values of the objective function at points suggested by `ask()`
        or other points. By default a new model will be fit to all
        observations. The new model is used to suggest the next point at
        which to evaluate the objective. This point can be retrieved by calling
        `ask()`.

        To add observations without fitting a new model set `fit` to False.

        To add multiple observations in a batch pass a list-of-lists for `x`
        and a list of scalars for `y`.

        Parameters
        ----------
        * `x` [list or list-of-lists]:
            Point at which objective was evaluated.

        * `y` [scalar or list]:
            Value of objective at `x`.

        * `fit` [bool, default=True]
            Fit a model to observed evaluations of the objective. A model will
            only be fitted after `n_initial_points` points have been told to
            the optimizer irrespective of the value of `fit`.
        """
        check_x_in_space(x, self.space)
        self._check_y_is_valid(x, y)

        # take the logarithm of the computation times
        if "ps" in self.acq_func:
            if is_2Dlistlike(x):
                y = [[val, log(t)] for (val, t) in y]
            elif is_listlike(x):
                y = list(y)
                y[1] = log(y[1])

        return self._tell(x, y, fit=fit)

    def _tell(self, x, y, fit=True):
        """Perform the actual work of incorporating one or more new points.
        See `tell()` for the full description.

        This method exists to give access to the internals of adding points
        by side stepping all input validation and transformation."""

        if "ps" in self.acq_func:
            if is_2Dlistlike(x):
                self.Xi.extend(x)
                self.yi.extend(y)
                self._n_initial_points -= len(y)
            elif is_listlike(x):
                self.Xi.append(x)
                self.yi.append(y)
                self._n_initial_points -= 1

        # If we have been handed a batch of multiobjective points
        elif self.n_objectives > 1 and is_2Dlistlike(x) and is_2Dlistlike(y):
            self.Xi.extend(x)
            self.yi.extend(y)
            self._n_initial_points -= len(y)

        # If we have been handed a single multiobjective point
        elif self.n_objectives > 1 and is_listlike(x) and is_listlike(y):
            self.Xi.append(x)
            self.yi.append(y)
            self._n_initial_points -= 1

        # if we have been handed a batch of single objective points
        elif is_listlike(y) and is_2Dlistlike(x) and self.n_objectives == 1:
            self.Xi.extend(x)
            self.yi.extend(y)
            self._n_initial_points -= len(y)

        # if we have been handed a single point with a single objective
        elif is_listlike(x) and self.n_objectives == 1:
            self.Xi.append(x)
            self.yi.append(y)
            self._n_initial_points -= 1
        else:
            raise ValueError(
                "Type of arguments `x` (%s) and `y` (%s) "
                "not compatible when number of objectives is (%s)."
                % (type(x), type(y), self.n_objectives)
            )

        # optimizer learned something new - discard cache
        self.cache_ = {}

        # after being "told" n_initial_points we switch from sampling
        # random points to using a surrogate model(s)
        if (
            fit
            and self._n_initial_points <= 0
            and self.base_estimator_ is not None
        ):
            transformed_bounds = np.array(self.space.transformed_bounds)
            est = clone(self.base_estimator_)

            # If the problem containts multiblie objectives a model has to be fitted for each objective
            if self.n_objectives > 1:

                # fit an estimator to each objective
                obj_models = []
                for i in range(self.n_objectives):
                    est = clone(self.base_estimator_)
                    y_list = [item[i] for item in self.yi]
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        est.fit(self.space.transform(self.Xi), y_list)
                    obj_models.append(est)

                # Append all objective functions
                self.models.append(obj_models)

                # Setting the probability of using Steinerberger for the next point (exploration)
                # The probability for using the NSGAII algorithm is 1-prob_stbr (exploitation)
                prob_stbr = 0.25

                # Simulate a random number
                random_uniform_number = np.random.uniform()

                # The random number decides what strategy to use for the next point
                if random_uniform_number < prob_stbr:

                    # self._next_x is found via stbr_scipy
                    next_x = self.stbr_scipy()
                    self._next_x = next_x[0]

                else:

                    # The Pareto front is approximated using the NSGAII algorithm
                    pop, logbook, front = self.NSGAII()

                    # The best point in the Pareto front is found (the point furthest from existing measurements)
                    next_x = self.best_Pareto_point(pop, front)
                    self._next_x = self.space.inverse_transform(
                        next_x.reshape((1, -1))
                    )[0]

            if self.n_objectives == 1:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    est.fit(self.space.transform(self.Xi), self.yi)

                if hasattr(self, "next_xs_") and self.acq_func == "gp_hedge":
                    self.gains_ -= est.predict(np.vstack(self.next_xs_))
                self.models.append(est)

                # even with BFGS as optimizer we want to sample a large number
                # of points and then pick the best ones as starting points
                if self._constraints:
                    # We use another sampling method if constraints have been added
                    X = self.space.transform(
                        self._constraints.rvs(
                            n_samples=self.n_points, random_state=self.rng
                        )
                    )
                else:
                    X = self.space.transform(
                        self.space.rvs(
                            n_samples=self.n_points, random_state=self.rng
                        )
                    )

                self.next_xs_ = []
                for cand_acq_func in self.cand_acq_funcs_:
                    values = _gaussian_acquisition(
                        X=X,
                        model=est,
                        y_opt=np.min(self.yi),
                        acq_func=cand_acq_func,
                        acq_func_kwargs=self.acq_func_kwargs,
                    )
                    # Find the minimum of the acquisition function by randomly
                    # sampling points from the space. If constraints are present
                    # we use this strategy
                    if self.acq_optimizer == "sampling" or self._constraints:
                        next_x = X[np.argmin(values)]

                    # Use BFGS to find the mimimum of the acquisition function, the
                    # minimization starts from `n_restarts_optimizer` different
                    # points and the best minimum is used
                    elif self.acq_optimizer == "lbfgs":
                        x0 = X[np.argsort(values)[: self.n_restarts_optimizer]]

                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore")
                            results = Parallel(n_jobs=self.n_jobs)(
                                delayed(fmin_l_bfgs_b)(
                                    gaussian_acquisition_1D,
                                    x,
                                    args=(
                                        est,
                                        np.min(self.yi),
                                        cand_acq_func,
                                        self.acq_func_kwargs,
                                    ),
                                    bounds=self.space.transformed_bounds,
                                    approx_grad=False,
                                    maxiter=20,
                                )
                                for x in x0
                            )

                        cand_xs = np.array([r[0] for r in results])
                        cand_acqs = np.array([r[1] for r in results])
                        next_x = cand_xs[np.argmin(cand_acqs)]

                    # lbfgs should handle this but just in case there are
                    # precision errors.
                    if not self.space.is_categorical:
                        next_x = np.clip(
                            next_x,
                            transformed_bounds[:, 0],
                            transformed_bounds[:, 1],
                        )
                    self.next_xs_.append(next_x)

                if self.acq_func == "gp_hedge":
                    logits = np.array(self.gains_)
                    logits -= np.max(logits)
                    exp_logits = np.exp(self.eta * logits)
                    probs = exp_logits / np.sum(exp_logits)
                    next_x = self.next_xs_[
                        np.argmax(self.rng.multinomial(1, probs))
                    ]
                else:
                    next_x = self.next_xs_[0]

                # note the need for [0] at the end
                self._next_x = self.space.inverse_transform(
                    next_x.reshape((1, -1))
                )[0]
        # Pack results

        return create_result(
            self.Xi, self.yi, self.space, self.rng, models=self.models
        )

    def _check_y_is_valid(self, x, y):
        """Check if the shape and types of x and y are consistent."""

        if "ps" in self.acq_func:
            if is_2Dlistlike(x):
                if not (np.ndim(y) == 2 and np.shape(y)[1] == 2):
                    raise TypeError("expected y to be a list of (func_val, t)")
            elif is_listlike(x):
                if not (np.ndim(y) == 1 and len(y) == 2):
                    raise TypeError("expected y to be (func_val, t)")

        # Check batch tell with multiobjective
        if is_2Dlistlike(x) and is_2Dlistlike(y) and self.n_objectives > 1:
            for y_values in y:
                if not len(y_values) == self.n_objectives:
                    raise ValueError(
                        "y does not have the correct number of objective scores"
                    )
                for y_value in y_values:
                    if not isinstance(y_value, Number):
                        raise ValueError(
                            "expected y to be a list of lists of scalars"
                        )

        # Check batch tell with single objective
        elif is_listlike(y) and is_2Dlistlike(x) and self.n_objectives == 1:
            for y_value in y:
                if not isinstance(y_value, Number):
                    raise ValueError("expected y to be a list of scalars")

        # Check single tell with multiobjective
        elif is_listlike(y):
            # Check if the observation has the correct number of objectives
            if not len(y) == self.n_objectives:
                raise ValueError(
                    "y does not have the correct number of objective scores"
                )
            # Check if all objective scores are numbers
            for y_value in y:
                if not isinstance(y_value, Number):
                    raise ValueError("expected y to be a list of scalars")

        # Check single tell with single objective
        elif is_listlike(x):
            if not isinstance(y, Number):
                raise ValueError("`func` should return a scalar")

        else:
            raise ValueError(
                "Type of arguments `x` (%s) and `y` (%s) "
                "not compatible." % (type(x), type(y))
            )

    def run(self, func, n_iter=1):
        """Execute ask() + tell() `n_iter` times"""
        for _ in range(n_iter):
            x = self.ask()
            self.tell(x, func(x))

        return create_result(
            self.Xi, self.yi, self.space, self.rng, models=self.models
        )

    def set_constraints(self, constraints):
        """ Sets the constraints for the optimizer

        Parameters
        ----------
        * `constraints` [list] or [Constraints]:
            Can either be a list of Constraint objects or a Constraints object
        """
        if self._n_initial_points > 0 and self._lhs:
            raise RuntimeError(
                "Can't set constraints while latin hypercube sampling points are not exhausted."
            )

        if constraints:
            if isinstance(constraints, Constraints):
                # If constraints is a Constraints object we simply add it
                self._constraints = constraints
            else:
                # If it is a list of constraints we initialize a Constraints object.
                self._constraints = Constraints(constraints, self.space)
        else:
            self._constraints = None

        self.update_next()

    def remove_constraints(self):
        """ Sets constraints to None"""
        self.set_constraints(None)

    def get_constraints(self):
        """Returns constraints"""
        return self._constraints

    def update_next(self):
        """ Updates the value returned by opt.ask(). Useful if a parameter was updated after ask was called."""
        self.cache_ = {}
        # Ask for a new next_x. Usefull if new constraints have been added or lenght_scale has been tweaked.
        # We only need to overwrite _next_x if it exists.
        if hasattr(self, "_next_x"):
            opt = self.copy(random_state=self.rng)
            self._next_x = opt._next_x

    def get_result(self):
        """Returns the same result that would be returned by opt.tell()
        but without calling tell.
        In the case of multiobejective optimization, a list of results are
        returned."""
        return create_result(
            self.Xi, self.yi, self.space, self.rng, models=self.models
        )

    def _check_length_scale_bounds(self, dimensions, bounds):
        """ Checks if length scale bounds are of in correct format"""
        space = Space(dimensions)
        n_dims = space.n_dims
        if bounds:
            if isinstance(bounds, list):
                # Check that number of bounds is the same as number of dimensions
                if not len(bounds) == n_dims:
                    raise ValueError(
                        "Number of bounds (%s) must be same as number of dimensions (%s)"
                        % (len(bounds), n_dims)
                    )
                for i in range(len(bounds)):
                    if not isinstance(bounds[i], tuple):
                        raise TypeError(
                            "Each bound must be of type tuple, got %s"
                            % (type(bounds[i]))
                        )
                    if not len(bounds[i]) == 2:
                        raise ValueError(
                            "Each bound-tuple must have length 2, got %s"
                            % (len(bounds[i]))
                        )
            else:
                raise TypeError(
                    "Expected bounds to be of type list, got %s"
                    % (type(bounds))
                )

    def stbr_scipy(self, n_points=1):
        from scipy.optimize import minimize

        # Suggestion for improvemenet. Seperate categorical and real/integer space.
        # Only solve Steinerberger minimization in real/integer sub space.

        # Get bounds for variables in transformed space
        # Needs to be a tuple of tuples with 0 and 1 for each transformed dimension
        bounds = []
        for i in range(self.space.transformed_n_dims):
            bounds.append((0.0, 1.0))
        bounds = tuple(tuple(sub) for sub in bounds)

        # Copy of the optimizer to not append new points to original  self.Xi
        # Set base estimator to GP so transform always nomalizes
        copy = Optimizer(
            self.space,
            "GP",
            n_objectives=self.n_objectives,
            n_initial_points=999,
        )
        for i in range(len(self.Xi)):
            if self.n_objectives == 1:
                copy.tell(self.Xi[i], 0)
            elif self.n_objectives > 1:
                copy.Xi.append(self.Xi[i])
                copy.yi.append(np.zeros(self.n_objectives).tolist())

        # Initialize list with Steinerberger points
        X = []
        # In each loop calculate the next Steinerberger point
        for i in range(n_points):
            # lists with local minimum location and the function value of the Steinerberger sum at that point
            loc_min = []
            fun_val = []
            # We use 20 lhs point as initial guesses for minimization
            x0 = copy.space.lhs(20)
            x0 = copy.space.transform(x0)

            # Loop over each initial guess and find a local minimum
            for j in range(len(x0)):
                res = minimize(copy.stbr_fun, x0=x0[j], bounds=bounds)
                loc_min.append(res.x)
                fun_val.append(res.fun)

            # Decide the global minimum as the local minimum with the lowest function value
            glob_min = loc_min[np.argmin(fun_val)]
            next_X = np.asarray(glob_min).reshape(
                1, copy.space.transformed_n_dims
            )
            # Transform back to original space
            next_X = copy.space.inverse_transform(next_X)

            # This deals with categorical variables, since they are not suited for Steinerberger
            for dim, n in zip(self.space.dimensions, range(self.space.n_dims)):
                # Check if dimension is categorical
                if isinstance(dim, Categorical):
                    # Make array with categories for that dimension
                    categories = dim.categories
                    # Make array with all instances of categories observed
                    instances = np.array(copy.Xi)[:, n]
                    # Calculate the number of instances of each category
                    cat_population = np.zeros(len(categories))
                    for category, i in zip(categories, range(len(categories))):
                        cat_population[i] = (instances.tolist()).count(
                            category
                        )
                    # Find the category with the lowest population
                    least_populated = categories[np.argmin(cat_population)]
                    # Set the category to the lowest populated category
                    next_X[0][n] = least_populated

            # Append point to list of new Steinerberger points
            X.append(next_X[0])
            # Append to Xi of copy optimizer
            copy.Xi.append(next_X[0])

        return X

    # This function returns the Steinerberger sum for a given x
    def stbr_fun(self, x):
        # parameter to ensure that log argument is non-zero
        eta = 10 ** -8
        # Initialize Steinerberger sum
        stbr_sum = 0
        # Transform initial points to [0,1]^d space
        Xi = self.space.transform(self.Xi)
        # for loop over all existing points
        for i in range(len(Xi)):
            # Calculate the factors in the Steinerberger term in each dimension
            stbr_vector = 1 - np.log(2 * np.sin(np.pi * abs(x - Xi[i])) + eta)
            # Calculate the Steinerberger term by multiplying factor from each dimension
            stbr_term = np.prod(stbr_vector)
            # Add term to Steinerberger sum
            stbr_sum = stbr_sum + stbr_term

        return stbr_sum

    # This function returns the objective scores at x estimated by the models
    def __ObjectiveGP(self, x):

        # Fator = 1.0e10
        F = [None] * self.n_objectives
        xx = np.asarray(x).reshape(1, -1)

        # Constraints = 0.0
        # for cons in self.constraints:
        #    y = cons['fun'](x)
        #    if cons['type'] == 'eq':
        #        Constraints += np.abs(y)
        #    elif cons['type'] == 'ineq':
        #        if y < 0:
        #            Constraints -= y

        for i in range(self.n_objectives):

            F[i] = self.models[-1][i].predict(xx)[0]

        return F

    # This function returns the point in the Pareto front, which is deemed the best (furthest away from existing observations)
    def best_Pareto_point(self, pop, front, q=0.5):

        Population = np.asarray(pop)

        IndexF, FatorF = self.__LargestOfLeast(front, self.yi)

        IndexPop, FatorPop = self.__LargestOfLeast(
            Population, self.space.transform(self.Xi).tolist()
        )

        Fator = q * FatorF + (1 - q) * FatorPop
        Index_try = np.argmax(Fator)

        best_point = Population[Index_try]

        return best_point

    # This function is used  in best_Pareto_point. For each point it gives a relative distance to the closest point
    def __LargestOfLeast(self, front, F):
        NF = len(front)
        MinDist = np.empty(NF)
        for i in range(NF):
            MinDist[i] = self.__MinimalDistance(front[i], F)
        ArgMax = np.argmax(MinDist)

        Mean = MinDist.mean()
        Std = np.std(MinDist)
        if (
            Std == 0
        ):  # Quick workaround to avoid divide by zero, when 0 or 1 datapoints. Investigate alternatives in future.
            return ArgMax, (MinDist - Mean)
        else:
            return ArgMax, (MinDist - Mean) / (Std)

    # This fuction returns the minimal distance between a point and a list of points
    @staticmethod
    def __MinimalDistance(X, Y):
        Y = np.asarray(Y)
        N = len(X)
        Npts = len(Y)
        DistMin = float("inf")
        for i in range(Npts):
            Dist = 0.0
            for j in range(N):
                Dist += (X[j] - Y[i, j]) ** 2
            Dist = np.sqrt(Dist)
            if Dist < DistMin:
                DistMin = Dist
        return DistMin

    # This function calls NSGAII to estimate the Pareto Front
    def NSGAII(self, MU=40):

        from ._NSGA2 import NSGAII

        pop, logbook, front = NSGAII(
            self.n_objectives,
            self.__ObjectiveGP,
            np.array(self.space.transformed_bounds),
            MU=MU,
        )

        return pop, logbook, front



"""Machine learning extensions for model-based optimization."""

from .forest import RandomForestRegressor
from .forest import ExtraTreesRegressor
from .gaussian_process import GaussianProcessRegressor
from .gbrt import GradientBoostingQuantileRegressor


__all__ = ("RandomForestRegressor",
           "ExtraTreesRegressor",
           "GradientBoostingQuantileRegressor",
           "GaussianProcessRegressor")



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='ProcessOptimizer',
      version='0.7.8',
      description='Sequential model-based optimization toolbox \
    (forked from scikit-optimize)',
      url='https://github.com/novonordisk-research/ProcessOptimizer',
      license='BSD',
      author='Novo Nordisk, Research & Early Development',
      packages=[
          'ProcessOptimizer',
          'ProcessOptimizer.learning',
          'ProcessOptimizer.model_systems',
          'ProcessOptimizer.optimizer',
          'ProcessOptimizer.space',
          'ProcessOptimizer.learning.gaussian_process'
          ],
      install_requires=['numpy', 'matplotlib', 'scipy', 'bokeh',
                        'scikit-learn>=0.24.2', 'six', 'deap', 'pyYAML'],
      extras_require={
          "browniebee": ['numpy==1.23.3',
                         'matplotlib==3.5.3',
                         'scipy==1.9.1',
                         'scikit-learn==1.1.2',
                         'six==1.16.0',
                         'deap==1.3.3',
                         'pyYAML==6.0',
                         'bokeh==2.4.3',
                         'tornado==6.2',
                         ]
          },
      long_description=long_description,
      long_description_content_type='text/markdown'
      )



import numpy as np
from ProcessOptimizer.model_systems import ModelSystem
from ProcessOptimizer.model_systems.benchmarks import branin
from ProcessOptimizer.space import Real

# Take the score function from the benchmarks file
# Add noise
def score(x, rng=np.random.default_rng(), noise_std=0.02): 
    return branin(x) + rng.normal(scale=noise_std)

# Define the relevant parameter space
space = []
space += [Real(-5,10,name='x1')]
space += [Real(0,15,name='x2')]

# Set the true minimum value
true_min = 0.397887

# Create model system
branin_hoo = ModelSystem(score, space, true_min)


import numpy as np
from ProcessOptimizer.model_systems import ModelSystem
from ProcessOptimizer.model_systems.benchmarks import poly_2d
from ProcessOptimizer.space import Real

# This system is intended to run in the domain x[0] in [-1,1] and x[1] in [-1,1]. This function
# has its minimum in this domain at (0.6667,-0.4833), with a value of 
# about -2.0512, and its maximum in this domain at (-1,-1) with a value of -1.27
def score(x, rng=np.random.default_rng(), noise_std=0.02):
    """Calculate the score of the model system

    Parameters:
    * x [array of length 2 containing numbers between -1 and 1]:
        The point in the parameter space at which to calculate the score. 
    * rng [np.random Generator]:
        Random number generator used to generate noise. 
    * noise_std [float]: 
        Standard deviation of the noise which is added to the result of the polynomial. 
    """
    value = poly_2d(x) + rng.normal(scale=noise_std)
    return value

# The location of the minimum in the parameter space:
true_min_loc = [2/3, -0.4833]
# The true minimum value
true_min = score(true_min_loc, noise_std=0)

# Define a the domain as a Space object
space = []
space += [Real(-1, 1, name='x1')]
space += [Real(-1, 1, name='x2')]

# Create model system
polynomial_system_2d = ModelSystem(score, space, true_min)


"""
Abstraction for optimizers.

It is sufficient that one re-implements the base estimator.
"""

import copy
import inspect
import numbers
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

from ..callbacks import check_callback
from ..callbacks import VerboseCallback
from .optimizer import Optimizer
from ..utils import eval_callbacks


def base_minimize(func, dimensions, base_estimator,
                  n_calls=100, n_random_starts=10,
                  acq_func="EI", acq_optimizer="lbfgs",
                  x0=None, y0=None, random_state=None, verbose=False,
                  callback=None, n_points=10000, n_restarts_optimizer=5,
                  xi=0.01, kappa=1.96, n_jobs=1):
    """
    Parameters
    ----------
    * `func` [callable]:
        Function to minimize. Should take a single list of parameters
        and return the objective value.

        If you have a search-space where all dimensions have names,
        then you can use `ProcessOptimizer.utils.use_named_args` as a decorator
        on your objective function, in order to call it directly
        with the named arguments. See `use_named_args` for an example.

    * `dimensions` [list, shape=(n_dims,)]:
        List of search space dimensions.
        Each search dimension can be defined either as

        - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
          dimensions),
        - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
          dimensions),
        - as a list of categories (for `Categorical` dimensions), or
        - an instance of a `Dimension` object (`Real`, `Integer` or
          `Categorical`).

         NOTE: The upper and lower bounds are inclusive for `Integer`
         dimensions.

    * `base_estimator` [sklearn regressor]:
        Should inherit from `sklearn.base.RegressorMixin`.
        In addition, should have an optional `return_std` argument,
        which returns `std(Y | x)`` along with `E[Y | x]`.

    * `n_calls` [int, default=100]:
        Maximum number of calls to `func`. An objective fucntion will
        always be evaluated this number of times; Various options to
        supply initialization points do not affect this value.

    * `n_random_starts` [int, default=10]:
        Number of evaluations of `func` with random points before
        approximating it with `base_estimator`.

    * `acq_func` [string, default=`"EI"`]:
        Function to minimize over the posterior distribution. Can be either

        - `"LCB"` for lower confidence bound,
        - `"EI"` for negative expected improvement,
        - `"PI"` for negative probability of improvement.
        - `"EIps" for negated expected improvement per second to take into
          account the function compute time. Then, the objective function is
          assumed to return two values, the first being the objective value and
          the second being the time taken in seconds.
        - `"PIps"` for negated probability of improvement per second. The
          return type of the objective function is assumed to be similar to
          that of `"EIps

    * `acq_optimizer` [string, `"sampling"` or `"lbfgs"`, default=`"lbfgs"`]:
        Method to minimize the acquistion function. The fit model
        is updated with the optimal value obtained by optimizing `acq_func`
        with `acq_optimizer`.

        - If set to `"sampling"`, then `acq_func` is optimized by computing
          `acq_func` at `n_points` randomly sampled points and the smallest
          value found is used.
        - If set to `"lbfgs"`, then
              - The `n_restarts_optimizer` no. of points which the acquisition
                function is least are taken as start points.
              - `"lbfgs"` is run for 20 iterations with these points as initial
                points to find local minima.
              - The optimal of these local minima is used to update the prior.

    * `x0` [list, list of lists or `None`]:
        Initial input points.

        - If it is a list of lists, use it as a list of input points. If no
          corresponding outputs `y0` are supplied, then len(x0) of total
          calls to the objective function will be spent evaluating the points
          in `x0`. If the corresponding outputs are provided, then they will
          be used together with evaluated points during a run of the algorithm
          to construct a surrogate.
        - If it is a list, use it as a single initial input point. The
          algorithm will spend 1 call to evaluate the initial point, if the
          outputs are not provided.
        - If it is `None`, no initial input points are used.

    * `y0` [list, scalar or `None`]
        Objective values at initial input points.

        - If it is a list, then it corresponds to evaluations of the function
          at each element of `x0` : the i-th element of `y0` corresponds
          to the function evaluated at the i-th element of `x0`.
        - If it is a scalar, then it corresponds to the evaluation of the
          function at `x0`.
        - If it is None and `x0` is provided, then the function is evaluated
          at each element of `x0`.

    * `random_state` [int, RandomState instance, or None (default)]:
        Set random state to something other than None for reproducible
        results.

    * `verbose` [boolean, default=False]:
        Control the verbosity. It is advised to set the verbosity to True
        for long optimization runs.

    * `callback` [callable, list of callables, optional]
        If callable then `callback(res)` is called after each call to `func`.
        If list of callables, then each callable in the list is called.

    * `n_points` [int, default=10000]:
        If `acq_optimizer` is set to `"sampling"`, then `acq_func` is
        optimized by computing `acq_func` at `n_points` randomly sampled
        points.

    * `n_restarts_optimizer` [int, default=5]:
        The number of restarts of the optimizer when `acq_optimizer`
        is `"lbfgs"`.

    * `xi` [float, default=0.01]:
        Controls how much improvement one wants over the previous best
        values. Used when the acquisition is either `"EI"` or `"PI"`.

    * `kappa` [float, default=1.96]:
        Controls how much of the variance in the predicted values should be
        taken into account. If set to be very high, then we are favouring
        exploration over exploitation and vice versa.
        Used when the acquisition is `"LCB"`.

    * `n_jobs` [int, default=1]:
        Number of cores to run in parallel while running the lbfgs
        optimizations over the acquisition function. Valid only when
        `acq_optimizer` is set to "lbfgs."
        Defaults to 1 core. If `n_jobs=-1`, then number of jobs is set
        to number of cores.

    Returns
    -------
    * `res` [`OptimizeResult`, scipy object]:
        The optimization result returned as a OptimizeResult object.
        Important attributes are:

        - `x` [list]: location of the minimum.
        - `fun` [float]: function value at the minimum.
        - `models`: surrogate models used for each iteration.
        - `x_iters` [list of lists]: location of function evaluation for each
           iteration.
        - `func_vals` [array]: function value for each iteration.
        - `space` [Space]: the optimization space.
        - `specs` [dict]`: the call specifications.
        - `rng` [RandomState instance]: State of the random state
           at the end of minimization.

        For more details related to the OptimizeResult object, refer
        http://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.OptimizeResult.html
    """
    specs = {"args": copy.copy(inspect.currentframe().f_locals),
             "function": inspect.currentframe().f_code.co_name}

    acq_optimizer_kwargs = {
        "n_points": n_points, "n_restarts_optimizer": n_restarts_optimizer,
        "n_jobs": n_jobs}
    acq_func_kwargs = {"xi": xi, "kappa": kappa}

    # Initialize optimization
    # Suppose there are points provided (x0 and y0), record them

    # check x0: list-like, requirement of minimal points
    if x0 is None:
        x0 = []
    elif not isinstance(x0[0], (list, tuple)):
        x0 = [x0]
    if not isinstance(x0, list):
        raise ValueError("`x0` should be a list, but got %s" % type(x0))
    if n_random_starts <= 0 and not x0:
        raise ValueError("Either set `n_random_starts` > 0,"
                         " or provide `x0`")
    # check y0: list-like, requirement of maximal calls
    if isinstance(y0, Iterable):
        y0 = list(y0)
    elif isinstance(y0, numbers.Number):
        y0 = [y0]
    required_calls = n_random_starts + (len(x0) if not y0 else 0)
    if n_calls < required_calls:
        raise ValueError(
            "Expected `n_calls` >= %d, got %d" % (required_calls, n_calls))
    # calculate the total number of initial points
    n_initial_points = n_random_starts + len(x0)

    # Build optimizer

    # create optimizer class
    optimizer = Optimizer(dimensions, base_estimator,
                          n_initial_points=n_initial_points,
                          acq_func=acq_func, acq_optimizer=acq_optimizer,
                          random_state=random_state,
                          acq_optimizer_kwargs=acq_optimizer_kwargs,
                          acq_func_kwargs=acq_func_kwargs)
    # check x0: element-wise data type, dimensionality
    assert all(isinstance(p, Iterable) for p in x0)
    if not all(len(p) == optimizer.space.n_dims for p in x0):
        raise RuntimeError("Optimization space (%s) and initial points in x0 "
                           "use inconsistent dimensions." % optimizer.space)
    # check callback
    callbacks = check_callback(callback)
    if verbose:
        callbacks.append(VerboseCallback(
            n_init=len(x0) if not y0 else 0,
            n_random=n_random_starts,
            n_total=n_calls))

    # Record provided points

    # create return object
    result = None
    # evaluate y0 if only x0 is provided
    if x0 and y0 is None:
        y0 = list(map(func, x0))
        n_calls -= len(y0)
    # record through tell function
    if x0:
        if not (isinstance(y0, Iterable) or isinstance(y0, numbers.Number)):
            raise ValueError(
                "`y0` should be an iterable or a scalar, got %s" % type(y0))
        if len(x0) != len(y0):
            raise ValueError("`x0` and `y0` should have the same length")
        result = optimizer.tell(x0, y0)
        result.specs = specs
        if eval_callbacks(callbacks, result):
            return result

    # Optimize
    for n in range(n_calls):
        next_x = optimizer.ask()
        next_y = func(next_x)
        result = optimizer.tell(next_x, next_y)
        result.specs = specs
        if eval_callbacks(callbacks, result):
            break

    return result



from copy import deepcopy
from functools import wraps

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import OptimizeResult
from scipy.optimize import minimize as sp_minimize
from sklearn.base import is_regressor
from sklearn.ensemble import GradientBoostingRegressor
from joblib import dump as dump_
from joblib import load as load_

from .learning import ExtraTreesRegressor
from .learning import GaussianProcessRegressor
from .learning import GradientBoostingQuantileRegressor
from .learning import RandomForestRegressor
from .learning.gaussian_process.kernels import ConstantKernel
from .learning.gaussian_process.kernels import HammingKernel
from .learning.gaussian_process.kernels import Matern

from .space import Space, Categorical, Integer, Real, Dimension

__all__ = (
    "load",
    "dump",
)


def create_result(Xi, yi, space=None, rng=None, specs=None, models=None):
    """
    Initialize an `OptimizeResult` object.

    Parameters
    ----------
    * `Xi` [list of lists, shape=(n_iters, n_features)]:
        Location of the minimum at every iteration.

    * `yi` [array-like, shape=(n_iters,)]:
        Minimum value obtained at every iteration.

    * `space` [Space instance, optional]:
        Search space.

    * `rng` [RandomState instance, optional]:
        State of the random state.

    * `specs` [dict, optional]:
        Call specifications.

    * `models` [list, optional]:
        List of fit surrogate models.

    Returns
    -------
    * `res` [`OptimizeResult`, scipy object]:
        OptimizeResult instance with the required information.

       or if the optimizer is multiobjective:
     * `results` [list of `OptimizeResult`, scipy object]:
        OptimizeResult instance with the required information.
    """
    res = OptimizeResult()
    yi = np.asarray(yi)
    if np.ndim(yi) == 1:
        best = np.argmin(yi)
        res.x = Xi[best]
        res.fun = yi[best]
        res.func_vals = yi
        res.x_iters = Xi
        res.models = models
        res.space = space
        res.random_state = rng
        res.specs = specs
        return res
    models = np.asarray(models)
    results = []
    for i in range(yi.shape[1]):
        res = OptimizeResult()
        yi_single = np.ravel(yi[:, i])
        best = np.argmin(yi_single)
        res.x = Xi[best]
        res.fun = yi_single[best]
        res.func_vals = yi_single
        res.x_iters = Xi
        if models.size == 0:
            res.models = models
        else:
            res.models = models[:, i]
        res.space = space
        res.random_state = rng
        res.specs = specs
        results.append(res)
    return results


def eval_callbacks(callbacks, result):
    """Evaluate list of callbacks on result.

    The return values of the `callbacks` are ORed together to give the
    overall decision on whether or not the optimization procedure should
    continue.

    Parameters
    ----------
    * `callbacks` [list of callables]:
        Callbacks to evaluate.

    * `result` [`OptimizeResult`, scipy object]:
        Optimization result object to be stored.

    Returns
    -------
    * `decision` [bool]:
        Decision of the callbacks whether or not to keep optimizing
    """
    stop = False
    if callbacks:
        for c in callbacks:
            decision = c(result)
            if decision is not None:
                stop = stop or decision

    return stop


def dump(res, filename, store_objective=True, **kwargs):
    """
    Store an ProcessOptimizer optimization result into a file.

    Parameters
    ----------
    * `res` [`OptimizeResult`, scipy object]:
        Optimization result object to be stored.

    * `filename` [string or `pathlib.Path`]:
        The path of the file in which it is to be stored. The compression
        method corresponding to one of the supported filename extensions ('.z',
        '.gz', '.bz2', '.xz' or '.lzma') will be used automatically.

    * `store_objective` [boolean, default=True]:
        Whether the objective function should be stored. Set `store_objective`
        to `False` if your objective function (`.specs['args']['func']`) is
        unserializable (i.e. if an exception is raised when trying to serialize
        the optimization result).

        Notice that if `store_objective` is set to `False`, a deep copy of the
        optimization result is created, potentially leading to performance
        problems if `res` is very large. If the objective function is not
        critical, one can delete it before calling `ProcessOptimizer.dump()`
        and thus avoid deep copying of `res`.

    * `**kwargs` [other keyword arguments]:
        All other keyword arguments will be passed to `joblib.dump`.
    """
    if store_objective:
        dump_(res, filename, **kwargs)

    elif 'func' in res.specs['args']:
        # If the user does not want to store the objective and it is indeed
        # present in the provided object, then create a deep copy of it and
        # remove the objective function before dumping it with joblib.dump.
        res_without_func = deepcopy(res)
        del res_without_func.specs['args']['func']
        dump_(res_without_func, filename, **kwargs)

    else:
        # If the user does not want to store the objective and it is already
        # missing in the provided object, dump it without copying.
        dump_(res, filename, **kwargs)


def load(filename, **kwargs):
    """
    Reconstruct a ProcessOptimizer optimization result from a file
    persisted with ProcessOptimizer.dump.

    Notice that the loaded optimization result can be missing
    the objective function (`.specs['args']['func']`) if
    `ProcessOptimizer.dump` was called with `store_objective=False`.

    Parameters
    ----------
    * `filename` [string or `pathlib.Path`]:
        The path of the file from which to load the optimization result.

    * `**kwargs` [other keyword arguments]:
        All other keyword arguments will be passed to `joblib.load`.

    Returns
    -------
    * `res` [`OptimizeResult`, scipy object]:
        Reconstructed OptimizeResult instance.
    """
    return load_(filename, **kwargs)


def is_listlike(x):
    return isinstance(x, (list, tuple))


def is_2Dlistlike(x):
    return np.all([is_listlike(xi) for xi in x])


def check_x_in_space(x, space):
    if is_2Dlistlike(x):
        if not np.all([p in space for p in x]):
            raise ValueError("Not all points are within the bounds of"
                             " the space.")
        if any([len(p) != len(space.dimensions) for p in x]):
            raise ValueError("Not all points have the same dimensions as"
                             " the space.")
    elif is_listlike(x):
        if x not in space:
            raise ValueError("Point (%s) is not within the bounds of"
                             " the space (%s)."
                             % (x, space.bounds))
        if len(x) != len(space.dimensions):
            raise ValueError(
                "Dimensions of point (%s) and space (%s) do not match"
                % (x, space.bounds)
                )


def expected_minimum(
                     res,
                     n_random_starts=20,
                     random_state=None,
                     return_std=False,
                     minmax='min'
                     ):
    """
    Compute the minimum over the predictions of the last surrogate model.

    Note that the returned minimum may not necessarily be an accurate
    prediction of the minimum of the true objective function.

    Parameters
    ----------
    * `res`  [`OptimizeResult`, scipy object]:
        The optimization result returned by a `ProcessOptimizer` minimizer.

    * `n_random_starts` [int, default=20]:
        The number of random starts for the minimization of the surrogate
        model.

    * 'return_std' [Boolean, default=True]:
        Whether the function should return the standard deviation or not.

    * `random_state` [int, RandomState instance, or None (default)]:
        Set random state to something other than None for reproducible
        results.

    * `minmax` [str, default='min']:
        Whether the function should return the expected minimun (intended use)
        or the expected maximum (edge use case).

    Returns
    -------
    * `x` [list]: location of the minimum (or maximum).

    * `fun` [float]: the surrogate function value at the minimum (or maximum).
    """
    if type(res) == list:
        raise ValueError(
                "expected_minimum does not support multiobjective results"
            )

    def func(x):
        reg = res.models[-1]
        if minmax == 'min':
            return reg.predict(x.reshape(1, -1))[0]
        elif minmax == 'max':
            return -1 * reg.predict(x.reshape(1, -1))[0]
        else:
            raise ValueError(
                    "expected minmax to be in ['min','max'], got %s"
                    % (minmax)
                )

    xs = [res.x]
    if n_random_starts > 0:
        xs.extend(res.space.rvs(n_random_starts, random_state=random_state))
    xs= res.space.transform(xs)
    best_x = None
    best_fun = np.inf

    for x0 in xs:
        r = sp_minimize(func, x0=x0, bounds=res.space.transformed_bounds)
        if r.fun < best_fun:
            best_x = res.space.inverse_transform(np.array(r.x).reshape(1,-1))[0]
            best_fun = r.fun

    if minmax == 'min':
        if return_std == True:
            std_estimate = res.models[-1].predict(res.space.transform([best_x]).reshape(1,-1), return_std=True)[1][0]
            return [v for v in best_x], [best_fun, std_estimate]
        else:
            return [v for v in best_x], best_fun


    elif minmax == 'max':
        if return_std == True:
            std_estimate = res.models[-1].predict(res.space.transform([best_x]).reshape(1,-1), return_std=True)[1][0]
            return [v for v in best_x], [-best_fun, std_estimate]
        else:
            return [v for v in best_x], -best_fun

    else:
        raise ValueError(
                "expected acq_func to be in ['min','max'], got %s"
                % (minmax)
            )


def expected_minimum_random_sampling(res,
                                     n_random_starts=100000,
                                     random_state=None,
                                     return_std=False,
                                     minmax='min'):
    """Minimum search by doing naive random sampling, Returns the parameters
    that gave the minimum function value. Can be used when the space
    contains any categorical values.
    .. note::
        The returned minimum may not necessarily be an accurate
        prediction of the minimum of the true objective function.
    Parameters
    ----------
    res : `OptimizeResult`, scipy object
        The optimization result returned by a `skopt` minimizer.
    n_random_starts : int, default=100000
        The number of random starts for the minimization of the surrogate
        model.
    random_state : int, RandomState instance, or None (default)
        Set random state to something other than None for reproducible
        results.
    Returns
    -------
    x : list]
        location of the minimum.
    fun : float
        the surrogate function value at the minimum.
    """

    # sample points from search space
    random_samples = res.space.rvs(n_random_starts, random_state=random_state)

    # make estimations with surrogate
    model = res.models[-1]
    y_random = model.predict(res.space.transform(random_samples))
    if minmax == 'min':
        index_best_objective = np.argmin(y_random)
    elif minmax == 'max':
        index_best_objective = np.argmax(y_random)
    else:
        raise ValueError(
                "expected minmax to be in ['min','max'], got %s"
                % (minmax)
            )

    extreme_x = random_samples[index_best_objective]
    if return_std == True:
        std_estimate = res.models[-1].predict(res.space.transform([extreme_x]).reshape(1,-1), return_std=True)[1][0]
        return extreme_x, [y_random[index_best_objective], std_estimate]
    else:
        return extreme_x, y_random[index_best_objective]


def has_gradients(estimator):
    """
    Check if an estimator's ``predict`` method provides gradients.

    Parameters
    ----------
    estimator: sklearn BaseEstimator instance.
    """
    tree_estimators = (
        ExtraTreesRegressor, RandomForestRegressor,
        GradientBoostingQuantileRegressor
    )

    # cook_estimator() returns None for "dummy minimize" aka random values only
    if estimator is None:
        return False

    if isinstance(estimator, tree_estimators):
        return False

    categorical_gp = False
    if hasattr(estimator, "kernel"):
        params = estimator.get_params()
        categorical_gp = (
            isinstance(estimator.kernel, HammingKernel) or
            any([isinstance(params[p], HammingKernel) for p in params])
        )

    return not categorical_gp


def cook_estimator(base_estimator,
                   space=None,
                   length_scale_bounds=None,
                   length_scale=None,
                   **kwargs):
    """
    Cook a default estimator.

    For the special base_estimator called "DUMMY" the return value is None.
    This corresponds to sampling points at random, hence there is no need
    for an estimator.

    Parameters
    ----------
    * `base_estimator` ["GP", "RF", "ET", "GBRT", "DUMMY"
                        or sklearn regressor, default="GP"]:
        Should inherit from `sklearn.base.RegressorMixin`.
        In addition the `predict` method should have an optional `return_std`
        argument, which returns `std(Y | x)`` along with `E[Y | x]`.
        If base_estimator is one of ["GP", "RF", "ET", "GBRT", "DUMMY"], a
        surrogate model corresponding to the relevant `X_minimize` function
        is created.

    * `space` [Space instance]:
        Has to be provided if the base_estimator is a gaussian process.
        Ignored otherwise.
    * `length_scale_bounds` [list of tuples]:
        the length scale bounds for the matern kernel
    * `length_scale_bounds` [list of floats]:
        the length scales for the Matern or Hamming kernel
    * `kwargs` [dict]:
        Extra parameters provided to the base_estimator at init time.
    """

    if isinstance(base_estimator, str):
        base_estimator = base_estimator.upper()
        if base_estimator not in ["GP", "ET", "RF", "GBRT", "DUMMY"]:
            raise ValueError("Valid strings for the base_estimator parameter "
                             " are: 'RF', 'ET', 'GP', 'GBRT' or 'DUMMY' not "
                             "%s." % base_estimator)
    elif not is_regressor(base_estimator):
        raise ValueError("base_estimator has to be a regressor.")

    if base_estimator == "GP":
        if space is not None:
            space = Space(space)
            space = Space(normalize_dimensions(space.dimensions))
            n_dims = space.transformed_n_dims
            is_cat = space.is_categorical

        else:
            raise ValueError("Expected a Space instance, not None.")

        cov_amplitude = ConstantKernel(1.0, (0.01, 1000))

        if not length_scale:
            length_scale = np.ones(n_dims)
        if not length_scale_bounds:
            length_scale_bounds = [(0.1, 1)] * n_dims

        # Transform lengthscale bounds:
        length_scale_bounds_transformed = []
        length_scale_transformed = []
        for i in range(len(space.dimensions)):
            for j in range(space.dimensions[i].transformed_size):
                length_scale_bounds_transformed.append(length_scale_bounds[i])
                length_scale_transformed.append(length_scale[i])

        # only special if *all* dimensions are categorical
        if is_cat:
            other_kernel = HammingKernel(length_scale=length_scale_transformed)
        else:
            other_kernel = Matern(
                length_scale=length_scale_transformed,
                length_scale_bounds=length_scale_bounds_transformed, nu=2.5)

        base_estimator = GaussianProcessRegressor(
            kernel=cov_amplitude * other_kernel,
            normalize_y=True, noise="gaussian",
            n_restarts_optimizer=4)
    elif base_estimator == "RF":
        base_estimator = RandomForestRegressor(n_estimators=100,
                                               min_samples_leaf=3)
    elif base_estimator == "ET":
        base_estimator = ExtraTreesRegressor(n_estimators=100,
                                             min_samples_leaf=3)
    elif base_estimator == "GBRT":
        gbrt = GradientBoostingRegressor(n_estimators=30, loss="quantile")
        base_estimator = GradientBoostingQuantileRegressor(base_estimator=gbrt)

    elif base_estimator == "DUMMY":
        return None

    base_estimator.set_params(**kwargs)
    return base_estimator


def dimensions_aslist(search_space):
    """Convert a dict representation of a search space into a list of
    dimensions, ordered by sorted(search_space.keys()).

    Parameters
    ----------
    search_space : dict
        Represents search space. The keys are dimension names (strings)
        and values are instances of classes that inherit from the class
        ProcessOptimizer.space.Dimension (Real, Integer or Categorical)
        Example:
            {'name1': Real(0,1), 'name2': Integer(2,4), 'name3': Real(-1,1)}

    Returns
    -------
    params_space_list: list of ProcessOptimizer.space.Dimension instances.
        Example output with example inputs:
            [Real(0,1), Integer(2,4), Real(-1,1)]
    """
    params_space_list = [
        search_space[k] for k in sorted(search_space.keys())
    ]
    return params_space_list


def point_asdict(search_space, point_as_list):
    """Convert the list representation of a point from a search space
    to the dictionary representation, where keys are dimension names
    and values are corresponding to the values of dimensions in the list.

    Counterpart to parameters_aslist.

    Parameters
    ----------
    search_space : dict
        Represents search space. The keys are dimension names (strings)
        and values are instances of classes that inherit from the class
        ProcessOptimizer.space.Dimension (Real, Integer or Categorical)
        Example:
            {'name1': Real(0,1), 'name2': Integer(2,4), 'name3': Real(-1,1)}

    point_as_list : list
        list with parameter values.The order of parameters in the list
        is given by sorted(params_space.keys()).
        Example:
            [0.66, 3, -0.15]

    Returns
    -------
    params_dict: dictionary with parameter names as keys to which
        corresponding parameter values are assigned.
        Example output with inputs:
            {'name1': 0.66, 'name2': 3, 'name3': -0.15}
    """
    params_dict = {
        k: v for k, v in zip(sorted(search_space.keys()), point_as_list)
    }
    return params_dict


def point_aslist(search_space, point_as_dict):
    """Convert a dictionary representation of a point from a search space to
    the list representation. The list of values is created from the values of
    the dictionary, sorted by the names of dimensions used as keys.

    Counterpart to parameters_asdict.

    Parameters
    ----------
    search_space : dict
        Represents search space. The keys are dimension names (strings)
        and values are instances of classes that inherit from the class
        ProcessOptimizer.space.Dimension (Real, Integer or Categorical)
        Example:
            {'name1': Real(0,1), 'name2': Integer(2,4), 'name3': Real(-1,1)}

    point_as_dict : dict
        dict with parameter names as keys to which corresponding
        parameter values are assigned.
        Example:
            {'name1': 0.66, 'name2': 3, 'name3': -0.15}

    Returns
    -------
    point_as_list: list with point values.The order of
        parameters in the list is given by sorted(params_space.keys()).
        Example output with example inputs:
            [0.66, 3, -0.15]
    """
    point_as_list = [
        point_as_dict[k] for k in sorted(search_space.keys())
    ]
    return point_as_list


def normalize_dimensions(dimensions):
    """Create a ``Space`` where all dimensions are normalized to unit range.

    This is particularly useful for Gaussian process based regressors and is
    used internally by ``gp_minimize``.

    Parameters
    ----------
    * `dimensions` [list, shape=(n_dims,)]:
        List of search space dimensions.
        Each search dimension can be defined either as

        - a `(lower_bound, upper_bound)` tuple (for `Real` or `Integer`
          dimensions),
        - a `(lower_bound, upper_bound, "prior")` tuple (for `Real`
          dimensions),
        - as a list of categories (for `Categorical` dimensions), or
        - an instance of a `Dimension` object (`Real`, `Integer` or
          `Categorical`).

         NOTE: The upper and lower bounds are inclusive for `Integer`
         dimensions.
    """
    space = Space(dimensions)
    transformed_dimensions = []
    if space.is_categorical:
        # recreate the space and explicitly set transform to "identity"
        # this is a special case for GP based regressors
        for dimension in space:
            transformed_dimensions.append(Categorical(dimension.categories,
                                                      dimension.prior,
                                                      name=dimension.name,
                                                      transform="identity"))

    else:
        for dimension in space.dimensions:
            if isinstance(dimension, Categorical):
                transformed_dimensions.append(dimension)
            # To make sure that GP operates in the [0, 1] space
            elif isinstance(dimension, Real):
                transformed_dimensions.append(
                    Real(dimension.low, dimension.high, dimension.prior,
                         name=dimension.name,
                         transform="normalize")
                )
            elif isinstance(dimension, Integer):
                transformed_dimensions.append(
                    Integer(dimension.low, dimension.high,
                            name=dimension.name,
                            transform="normalize")
                )
            else:
                raise RuntimeError("Unknown dimension type "
                                   "(%s)" % type(dimension))

    return Space(transformed_dimensions)


def use_named_args(dimensions):
    """
    Wrapper / decorator for an objective function that uses named arguments
    to make it compatible with optimizers that use a single list of parameters.

    Your objective function can be defined as being callable using named
    arguments: `func(foo=123, bar=3.0, baz='hello')` for a search-space
    with dimensions named `['foo', 'bar', 'baz']`. But the optimizer
    will only pass a single list `x` of unnamed arguments when calling
    the objective function: `func(x=[123, 3.0, 'hello'])`. This wrapper
    converts your objective function with named arguments into one that
    accepts a list as argument, while doing the conversion automatically.

    The advantage of this is that you don't have to unpack the list of
    arguments `x` yourself, which makes the code easier to read and
    also reduces the risk of bugs if you change the number of dimensions
    or their order in the search-space.

    Example Usage
    -------------
    # Define the search-space dimensions. They must all have names!
    dim1 = Real(name='foo', low=0.0, high=1.0)
    dim2 = Real(name='bar', low=0.0, high=1.0)
    dim3 = Real(name='baz', low=0.0, high=1.0)

    # Gather the search-space dimensions in a list.
    dimensions = [dim1, dim2, dim3]

    # Define the objective function with named arguments
    # and use this function-decorator to specify the search-space dimensions.
    @use_named_args(dimensions=dimensions)
    def my_objective_function(foo, bar, baz):
        return foo ** 2 + bar ** 4 + baz ** 8

    # Now the function is callable from the outside as
    # `my_objective_function(x)` where `x` is a list of unnamed arguments,
    # which then wraps your objective function that is callable as
    # `my_objective_function(foo, bar, baz)`.
    # The conversion from a list `x` to named parameters `foo`, `bar`, `baz`
    # is done automatically.

    # Run the optimizer on the wrapped objective function which is called as
    # `my_objective_function(x)` as expected by `forest_minimize()`.
    result = forest_minimize(func=my_objective_function, dimensions=dimensions,
                             n_calls=20, base_estimator="ET", random_state=4)

    # Print the best-found results.
    print("Best fitness:", result.fun)
    print("Best parameters:", result.x)

    Parameters
    ----------
    * `dimensions` [list(Dimension)]:
        List of `Dimension`-objects for the search-space dimensions.

    Returns
    -------
    * `wrapped_func` [callable]
        Wrapped objective function.
    """

    def decorator(func):
        """
        This uses more advanced Python features to wrap `func` using a
        function-decorator, which are not explained so well in the
        official Python documentation.

        A good video tutorial explaining how this works is found here:
        https://www.youtube.com/watch?v=KlBPCzcQNU8

        Parameters
        ----------
        * `func` [callable]:
            Function to minimize. Should take *named arguments*
            and return the objective value.
        """

        # Ensure all dimensions are correctly typed.
        if not all(isinstance(dim, Dimension) for dim in dimensions):
            # List of the dimensions that are incorrectly typed.
            err_dims = list(filter(lambda dim: not isinstance(dim, Dimension),
                                   dimensions))

            # Error message.
            msg = '''All dimensions must be instances of the Dimension-class,
                  but found: {}'''
            msg = msg.format(err_dims)
            raise ValueError(msg)

        # Ensure all dimensions have names.
        if any(dim.name is None for dim in dimensions):
            # List of the dimensions that have no names.
            err_dims = list(filter(lambda dim: dim.name is None, dimensions))

            # Error message.
            msg = "All dimensions must have names, but found: {}"
            msg = msg.format(err_dims)
            raise ValueError(msg)

        @wraps(func)
        def wrapper(x):
            """
            This is the code that will be executed every time the
            wrapped / decorated `func` is being called.
            It takes `x` as a single list of parameters and
            converts them to named arguments and calls `func` with them.

            Parameters
            ----------
            * `x` [list]:
                A single list of parameters e.g. `[123, 3.0, 'linear']`
                which will be converted to named arguments and passed
                to `func`.

            Returns
            -------
            * `objective_value`
                The objective value returned by `func`.
            """

            # Ensure the number of dimensions match
            # the number of parameters in the list x.
            if len(x) != len(dimensions):
                msg = "Mismatch in number of search-space dimensions. " \
                      "len(dimensions)=={} and len(x)=={}"
                msg = msg.format(len(dimensions), len(x))
                raise ValueError(msg)

            # Create a dict where the keys are the names of the dimensions
            # and the values are taken from the list of parameters x.
            arg_dict = {dim.name: value for dim, value in zip(dimensions, x)}

            # Call the wrapped objective function with the named arguments.
            objective_value = func(**arg_dict)

            return objective_value

        return wrapper

    return decorator


def y_coverage(res, return_plot=False, random_state=None, horizontal=False):
    """
    A function to calculate the expected range of observable function values
    given a result instans. This can be compared with the actual observed
    range of function values.
    """
    assert len(res.func_vals) != 0, "train model before using this function"
    observed_min = res.func_vals.min()
    observed_max = res.func_vals.max()
    min_x, expected_min = expected_minimum(res,
                                           n_random_starts=20,
                                           random_state=None,
                                           minmax='min')
    max_x, expected_max = expected_minimum(res,
                                           n_random_starts=20,
                                           random_state=None,
                                           minmax='max')

    if return_plot:
        reg = res.models[-1]
        min_x = res.space.transform([min_x, ])
        max_x = res.space.transform([max_x, ])
        sampled_mins = reg.sample_y(min_x,
                                    n_samples=5000,
                                    random_state=random_state)[0]
        sampled_maxs = reg.sample_y(max_x,
                                    n_samples=5000,
                                    random_state=random_state)[0]
        extreme_min = sampled_mins.min()
        extreme_max = sampled_maxs.max()
        bins = np.linspace(extreme_min, extreme_max, 30)
        colors = ['#B8DE29FF', '#453781FF']

        if horizontal:
            fig, ax = plt.subplots()
            ax.hist([
                sampled_mins,
                sampled_maxs],
                bins,
                label=['expected min', 'expected max'],
                orientation='horizontal',
                density=True,
                color=colors)
            ax.set_xlabel('"Plausibility" of achieving/realizing '
                          'given function value')
            ax.set_ylabel('Function Value')
            for i in range(len(res.func_vals)):
                if i == 0:
                    ax.axhline(y=res.func_vals[i],
                               xmin=0.6,
                               xmax=1,
                               color="darkorange",
                               alpha=0.5,
                               label='Observed points')
                else:
                    ax.axhline(y=res.func_vals[i],
                               xmin=0.6,
                               xmax=1,
                               color="darkorange",
                               alpha=0.5)
            ax.legend(loc='best', shadow=True)
            ax.set_xticks([])
            plt.show()

        else:
            fig, ax = plt.subplots()
            ax.hist([
                sampled_mins,
                sampled_maxs],
                bins,
                label=['expected min', 'expected max'],
                density=True,
                color=colors)
            ax.set_xlabel('Function value')
            ax.set_ylabel('"Plausibility" of achieving/realizing '
                          'given function value')
            for i in range(len(res.func_vals)):
                if i == 0:
                    ax.axvline(x=res.func_vals[i],
                               ymin=0.3,
                               ymax=0.7,
                               color="darkorange",
                               alpha=0.5,
                               label='Observed points')
                else:
                    ax.axvline(x=res.func_vals[i],
                               ymin=0.3,
                               ymax=0.7,
                               color="darkorange",
                               alpha=0.5)
            ax.legend(loc='best', shadow=True)
            ax.set_yticks([])
            plt.show()

    return (observed_min, observed_max), (expected_min, expected_max)
